-- Generating time-LTS (alpha^1) --

----------------------------------------------------------------------


-- | Defines the SysTick period
t0 :: Int
t0 = 4

----------------------------------------------------------------------

-- | Creates the initial priority queue.
--   Each pair contains the agent number and its prioryty.
q0 :: [(Int,Int)]
q0 = createPriorityQueue s0 1 modelSize []

----------------------------------------------------------------------

-- | This is auxiliary recursive function for creating initial priority queue.
createPriorityQueue :: State          -- ^ model state
                    -> Int            -- ^ number of the agent
                    -> Int            -- ^ model size (number of agents)
                    -> [(Int,Int)]    -- ^ temporary queue
                    -> [(Int,Int)]    -- ^ return value
createPriorityQueue state k n queue
  | takeam k state == X && k < n
    = createPriorityQueue state (k+1) n (insertIntoQueue (k, agentPriority (agents !! (k-1))) queue)
  | takeam k state == X && k == n = insertIntoQueue (k, agentPriority (agents !! (k-1))) queue
  | takeam k state /= X && k < n = createPriorityQueue state (k+1) n queue
  | otherwise = queue

----------------------------------------------------------------------

-- | Inserts elements into priority queue preserving the priority order.
insertIntoQueue :: (Int,Int)           -- ^ new pair to insert (agent number, priority)
                -> [(Int,Int)]         -- ^ current queue
                -> [(Int,Int)]         -- ^ return value
insertIntoQueue (x,xp) [] = [(x,xp)]
insertIntoQueue (x,xp) ((y,yp):ys)
   | x == y    = ((y,yp):ys)
   | xp < yp   = (x,xp):(y,yp):ys
   | otherwise = (y,yp) : (insertIntoQueue (x,xp) ys)

----------------------------------------------------------------------

-- | Type that represents a node of the time LTS graph (alpha^1 system layer).
-- (LTS_node_number, State, Priority queue, Time from SysTick, [outgoing_arcs])
-- outgoing arc: transition, time, destination node number
type Node = (Int, State, [(Int,Int)], Int, [(TTransition, Int, Int)])

----------------------------------------------------------------------

-- | Returns True if the argument is CLock entry, returns False otherwise.
isLock :: ContextInfo -> Bool
isLock (CLock _) = True
isLock _ = False

----------------------------------------------------------------------

-- | Returns True if the argument contains a CLock entry, returns False otherwise.
containsLock :: Set ContextInfo -> Bool
containsLock ci = (Set.size (Set.filter isLock ci)) > 0

----------------------------------------------------------------------

-- | Removes a CLock entry from the given context information list (set).
removeLock :: Set ContextInfo -> Set ContextInfo
removeLock ci = difference ci (Set.filter isLock ci)

----------------------------------------------------------------------

-- | Returns the argument of the CLock entry.
lockName :: ContextInfo -> Agent
lockName (CLock a) = a

----------------------------------------------------------------------

-- | Returns the name of agent that locked the given agent (second argument).
--   If the agent is not locked, its own name is returned.
lockingAgent :: State -> Agent -> Agent
lockingAgent state agent
  | lock == empty = agent
  | otherwise = lockName (head (toList lock))
  where n  = agentNumber agent
        ci = takeci n state
        lock = Set.filter isLock ci

----------------------------------------------------------------------

-- | Returns the name of agent that is locked by the given agent (second argument).
--   If none agent is locked, the second argument is returned.
lockedAgent :: State -> Agent -> Agent
lockedAgent state agent
  | locked == [] = agent
  | otherwise = head locked   -- there is exactly one
  where locked = [a | (a,b) <- zip agents (Prelude.map (lockingAgent state) agents), b == agent, a /= agent]

----------------------------------------------------------------------

-- | Searches for the active agent that is context agent for the given passive one.
-- | Returns None if such an agent does not exists.
contextAgent :: Int     -- ^ number of the agent
             -> State   -- ^ model state
             -> Agent   -- ^ return value
contextAgent k state
  | mode /= T = agents !! (k-1)
  | takeam n state == X = agents !! (n-1)
  | otherwise = contextAgent n state    -- n-th mode must be T
  where mode = takeam k state
        n = searchProc k 1 state        -- number of agent that called procedure of k-th agent
  
----------------------------------------------------------------------

-- | Provides the list of all transitions enable in the given state 
--   based on the alpha^1 system layers rules.
alpha1EnableTransitions :: State            -- ^ current state
                        -> [(Int,Int)]      -- ^ current priority queue
                        -> Int              -- ^ time to sysTick
                        -> [TTransition]    -- ^ active transitions
alpha1EnableTransitions state queue time 
  | queue == [] = []                                              -- terminal state
  | time == 0 = [STSysTick]
  | enableTrans == [] && agShifts == [-1] = []                    -- terminal state
  | systemTrans /= [] = systemTrans
  | agent1EnableTrans == [] && agent2EnableTrans == [] && agShifts == [-1] = [STTime time]    -- shift to the nearest sysTick
  | agent1EnableTrans == [] && agent2EnableTrans == [] = [STTime (min time (head agShifts))]  -- agShifts /= [-1]
  | agent2EnableTrans == [] = agent1EnableTrans     -- agent1EnableTrans /= []
  | agent1EnableTrans == [] = agent2EnableTrans     -- agent1EnableTrans /= []
  where enableTrans = enableTransitions state                         -- enable transitions for all X and R agents
        systemTrans = selectSystemTransitions enableTrans
        agent1 = agents !! ((fst (head queue))-1)
        passiveAgents = [a | a <- agents, a /= agent1, contextAgent (agentNumber a) state == agent1, procFree (takeci (agentNumber a) state)]  -- at most one
        agent2 = head passiveAgents
        agent1EnableTrans = selectTransitions agent1 enableTrans          -- enable transitions for the X (running) agent
        agent2EnableTrans = if passiveAgents /= [] then (selectTransitions agent2 enableTrans) else []  -- enable transitions for the passive agent
        agShifts = agentsShifts modelSize state

----------------------------------------------------------------------

-- | Selects transitions of the given agent from the given set of transitions. 
selectTransitions :: Agent -> [TTransition] -> [TTransition]
selectTransitions _ [] = []
selectTransitions agent (tr:trs) 
  | agent == transAgent tr = tr : (selectTransitions agent trs)
  | otherwise = selectTransitions agent trs

----------------------------------------------------------------------

-- | Selects system transitions from the given set of transitions. 
selectSystemTransitions :: [TTransition] -> [TTransition]
selectSystemTransitions [] = []
selectSystemTransitions (tr:trs) 
  | isSystemTransition tr = tr : (selectSystemTransitions trs)
  | otherwise = selectSystemTransitions trs

----------------------------------------------------------------------

-- | Returns the duration for system transitions.
systemTransitionsDuration :: Int
systemTransitionsDuration = 0

----------------------------------------------------------------------

-- | Returns the duration for given transition.
transDuration :: TTransition -> Int
transDuration (STInAP   _ _ _) = systemTransitionsDuration
transDuration (STInPP   _ _ _) = systemTransitionsDuration
transDuration (STOutAP  _ _ _) = systemTransitionsDuration
transDuration (STOutPP  _ _ _) = systemTransitionsDuration
transDuration (STDelayEnd _ _) = systemTransitionsDuration
transDuration (STLoopEnd  _ _) = systemTransitionsDuration
transDuration (STInEnd    _ _) = systemTransitionsDuration
transDuration (STOutEnd   _ _) = systemTransitionsDuration
transDuration (STSysTick     ) = systemTransitionsDuration
transDuration (STTime d)       = d
transDuration tr = duration (transAgent tr) (transNumber tr)

----------------------------------------------------------------------

-- | Returns list of integers used to determine time shift for the given agent.
--   This is auxiliary function. The function ignores CSFT entries.
ci2time :: Mode           -- ^ Agent mode
        -> Int            -- ^ Program counter
        -> [ContextInfo]  -- ^ Context info list
        -> [Int]          -- ^ List of time shifts
ci2time am pc ((CTimer _ d):cis)  = d : (ci2time am pc cis)
ci2time am pc ((CTimeout n):cis)
  | am == W && pc == n = [0]
  | otherwise = ci2time am pc cis
ci2time am pc (_:cis) = ci2time am pc cis
ci2time _ _ [] = []

----------------------------------------------------------------------

-- | Returns list of ciShifts for selected agents in the given state.
--   The list is reduced to the important values only (e.g. the list contains
--   at most one value equal to -1 as its last element).
agentsShifts :: Int    -- ^ Number of agents (modelSize)
             -> State  -- ^ Current state
             -> [Int]  -- ^ importatnt ciShifts
agentsShifts 1 state
  | shiftList == [] = [-1]
  | otherwise = [Prelude.minimum shiftList]
  where shiftList = ci2time (takeam 1 state) (takepc 1 state) (toList (takeci 1 state))

agentsShifts n state
  | shiftList == [] = agentsShifts (n - 1) state
  | listMin == 0    = [0]
  | otherwise       = listMin : (agentsShifts (n - 1) state)
  where shiftList = ci2time (takeam n state) (takepc n state) (toList (takeci n state))
        listMin   = Prelude.minimum shiftList

----------------------------------------------------------------------

-- | Returns the shift time for the given state and transition.
timeShift :: State           -- ^ current state
          -> TTransition     -- ^ current transition
          -> Int             -- ^ time to sysTick
          -> Int             -- ^ time shift for the considered transition
timeShift _ STSysTick _  = 0
timeShift _ (STTime d) _ = d
timeShift state tr time
  | agShifts == [-1] && sft == -1 = min trShift time 
  | agShifts == [-1] = Prelude.minimum [trShift, time, sft]
  | sft == -1 = Prelude.minimum [trShift, time, minShift]
  | otherwise = Prelude.minimum [trShift, time, sft, minShift]
  where agShifts = agentsShifts modelSize state
        minShift = minimum (Prelude.filter (>= 0) agShifts)
        trShift  = transDuration tr
        sft      = agentSFT (transAgent tr) state

----------------------------------------------------------------------

-- | Updates a context CTimer entry due to the passage of time.
updateTimer :: Int          -- ^ passage of time
            -> ContextInfo  -- ^ context entry
            -> ContextInfo  -- ^ udpated entry
updateTimer delta (CTimer n d)
  | delta == d = CTimeout n
  | delta < d  = CTimer n (d - delta)
updateTimer _ entry = entry

----------------------------------------------------------------------

-- | Updates a context information list due to the passage of time (CTimer
--   entries only).
updateAllCITimers :: Int -> Set ContextInfo -> Set ContextInfo
updateAllCITimers delta ci = Set.map (updateTimer delta) ci

----------------------------------------------------------------------

-- | Updates context information lists due to the passage of time (CTimer
--   entries only).
updateAllTimers :: Int -> State -> State
updateAllTimers delta state = updateAllContextInfo (updateAllCITimers delta) state

----------------------------------------------------------------------

-- | Returns True if the argument is CSFT entry, returns False otherwise.
isSFT :: ContextInfo -> Bool
isSFT (CSFT _) = True
isSFT _ = False

----------------------------------------------------------------------

-- | Returns the time argument for the given entry or -1 it there is no such an argument.
contextInfo2Time :: ContextInfo -> Int
contextInfo2Time (CTimer _ d)  = d
contextInfo2Time (COTimer _ d) = d
contextInfo2Time (CTimeout _)  = 0
contextInfo2Time (CSFT d)      = d
contextInfo2Time _             = -1

----------------------------------------------------------------------

-- | Returns the agent step-finish-time or -1 if none transition of
--   the agent is in progress in the given state.
agentSFT :: Agent -> State -> Int
agentSFT agent state
  | sftEntry == empty = -1
  | otherwise         = sft
  where n  = agentNumber agent
        ci = takeci n state
        sftEntry = Set.filter (isSFT) ci
        sft = contextInfo2Time (head (elems sftEntry))

----------------------------------------------------------------------

-- | Returns True if firing of the given transition cannot influence
--   activity of other agents and False otherwise.
isLocalTransition :: TTransition -> Bool
isLocalTransition (TDelay     _ _ )  = True
isLocalTransition (TExec      _ _ )  = True
isLocalTransition (TExit      _ _ )  = True
isLocalTransition (TIn        _ _  ) = True
isLocalTransition (TInAP      _ _ _) = True
isLocalTransition (TInPP      _ _ _) = True
isLocalTransition (TJump      _ _ )  = True
isLocalTransition (TLoop      _ _ )  = True
isLocalTransition (TLoopEvery _ _ )  = True
isLocalTransition (TNull      _ _ )  = True
isLocalTransition (TOut       _ _  ) = True
isLocalTransition (TOutAP     _ _ _) = True
isLocalTransition (TOutPP     _ _ _) = True
isLocalTransition (TSelect    _ _ )  = True
isLocalTransition _ = False
-- isLocalTransition (STInAP     _ _)   = False
-- isLocalTransition (STInPP     _ _)   = False
-- isLocalTransition (STOutAP    _ _)   = False
-- isLocalTransition (STOutPP    _ _)   = False
-- isLocalTransition (STDelayEnd _ _ )  = False
-- isLocalTransition (STLoopEnd  _ _ )  = False
-- isLocalTransition (STInEnd    _ _  ) = False
-- isLocalTransition (STOutEnd   _ _  ) = False 
-- isLocalTransition (TInF       _ _ _) = False
-- isLocalTransition (TOutF      _ _ _) = False
-- isLocalTransition (TStart     _ _ )  = False

----------------------------------------------------------------------

-- | Updates priority queue using the update state (next state).
updateQueue :: [(Int,Int)]    -- ^ current queue
            -> State          -- ^ model state
            -> [(Int,Int)]    -- ^ return value
updateQueue queue state = updateQueueWithAgents queue aps
  where as = Prelude.map transAgent (enableTransitions state)
        aps = [(agentNumber (contextAgent (agentNumber a) state), agentPriority (contextAgent (agentNumber a) state)) | a <- as]

----------------------------------------------------------------------

-- | Insert a list of agents into an existing priority queue.
updateQueueWithAgents :: [(Int,Int)]   -- ^ current queue
                      -> [(Int,Int)]   -- ^ pairs (agent number, prioroty) for ready agents
                      -> [(Int,Int)]   -- ^ return value
updateQueueWithAgents queue [] = queue 
updateQueueWithAgents queue (a:as) 
  = updateQueueWithAgents (insertIntoQueue a queue) as

----------------------------------------------------------------------

-- | Adds a context information to the context list (set) of the n-th agent.
addContextInfo :: Int -> ContextInfo -> State -> State
addContextInfo n info = updateContextInfo n (insert info)

----------------------------------------------------------------------

-- | Removes a context information from the context list (set) of the n-th agent.
delContextInfo :: Int -> ContextInfo -> State -> State
delContextInfo n info = updateContextInfo n (delete info)

----------------------------------------------------------------------

-- | Returns the CTimer, CTimeout or COTimer entry related to given statement number.
selectTimer :: Int -> Set ContextInfo -> ContextInfo
selectTimer n ci = selectTimerFromList n (toList ci)

----------------------------------------------------------------------

-- | Returns the CTimer, CTimeout or COTimer entry related to given statement number.
--   This is auxiliary funtion that searches a list of entries.

selectTimerFromList :: Int -> [ContextInfo] -> ContextInfo
selectTimerFromList _ [] = CNone
selectTimerFromList n ((CTimer k d):cis)
  | n == k = CTimer k d
  | otherwise = selectTimerFromList n cis
selectTimerFromList n ((CTimeout k):cis)
  | n == k = CTimeout k
  | otherwise = selectTimerFromList n cis
selectTimerFromList n ((COTimer k d):cis)
  | n == k = COTimer k d
  | otherwise = selectTimerFromList n cis
selectTimerFromList n (_:cis) = selectTimerFromList n cis

----------------------------------------------------------------------

-- | Generates the list of succeeding states for the given state and transition
--   fired in the state. This function takes the shift time into consideration
--   and generates only partial results. This is auxiliary function.
timeFire :: Int                        -- ^ time shift 
         -> TTransition                -- ^ transition
         -> State                      -- ^ current state
         -> [(Int,Int)]                -- ^ current queue
         -> Int                        -- ^ time to sysTick
         -> [(State,[(Int,Int)],Int)]  -- ^ new states (state, queue, sysTick time)
  
timeFire _ STSysTick state queue _ 
  | length queue == 1 = [(state,queue,t0)]
  | agent1EnableTrans /= [] || agent2EnableTrans /= [] 
    = [(state,(head (tail queue)):(insertIntoQueue (n1, agentPriority agent1) (tail (tail queue))),t0)]
  | otherwise = [(state,(tail queue),t0)]
  where n1               = fst (head queue)
        agent1           = agents !! (n1-1)
        enableTrans      = enableTransitions state
        agent1EnableTrans = selectTransitions agent1 enableTrans
        passiveAgents = [a | a <- agents, a /= agent1, contextAgent (agentNumber a) state == agent1, procFree (takeci (agentNumber a) state)]  -- at most one
        agent2 = head passiveAgents
        agent2EnableTrans = if passiveAgents /= [] then (selectTransitions agent2 enableTrans) else []  -- enable transitions for the passive agent

timeFire _ (STTime d) state queue time = [((updateAllTimers d state), queue, time-d)]

timeFire delta tr state queue time 
-- 1 -- fire transition, nothing to do with sft and timers
--
  | isLocalTransition tr == True && trTime == 0 = [(s,queue,time) | s <- (fire tr state)]
--
-- 2 -- fire transition, nothing to do with sft and timers, we must update queue
--
  | trTime == 0 = [(s,(updateQueue queue s),time) | s <- (fire tr state)]
--
-- 3 -- local transtion, no sft entry, we can fire complete transition
--
  | isLocalTransition tr == True && sft == -1 && trTime == delta 
    = [(s,queue,time-delta) | s <- (fire tr (updateAllTimers delta state))]
--
-- 4 -- local transtion, there is sft, we can finish the transition
--
  | isLocalTransition tr == True && sft > 0 && sft == delta 
    = [(s,queue,time-delta) | s <- (fire tr (updateAllTimers delta (delContextInfo n1 (CSFT sft) state)))]
--
-- 5 -- local transtion, no sft, we cannot complete the transition
--
  | isLocalTransition tr == True && sft == -1 && trTime > delta 
    = [(updateAllTimers delta (addContextInfo n1 (CSFT (trTime-delta)) state),queue,time-delta)]
--
-- 6 -- local transtion, there is sft, we cannot complete the transition
--
  | isLocalTransition tr == True && sft > 0 && sft < delta 
    = [(updateAllTimers delta (addContextInfo n1 (CSFT (sft-delta)) (delContextInfo n1 (CSFT sft) state)),queue,time-delta)]
--
-- 3' -- non-local transition, no sft entry, we can fire complete transition, we must update queue
--
  | sft == -1 && trTime == delta
    = [(s,(updateQueue queue s),time-delta) | s <- (fire tr (updateAllTimers delta state))]
--
-- 4' -- non-local agent-in-out transition, there is sft, we can finish the transition, we must update queue, we must remove CLock
--
  | isAgentInOutTransition tr && sft > 0 && sft == delta
    = [(s,(updateQueue queue s),time-delta) | s <- (fire tr (updateAllTimers delta (delContextInfo n2 (CLock agent1) (delContextInfo n1 (CSFT sft) state))))]
--
-- 4" -- non-local non-agent-in-out transition, there is sft, we can finish the transition, we must update queue
--    
  | sft > 0 && sft == delta
    = [(s,(updateQueue queue s),time-delta) | s <- (fire tr (updateAllTimers delta (delContextInfo n1 (CSFT sft) state)))]
--
-- 5' -- non-local agent-in-out transition, no sft, but we cannot complete the transition, we must update queue, we must add CLock
--
  | isAgentInOutTransition tr && sft == -1 && trTime > delta
    = [(state5prime,(updateQueue queue state5prime),time-delta)]
--
-- 5" -- non-local non-agent-in-out transition, no sft, but we cannot complete the transition, we must update queue
--    
  | sft == -1 && trTime > delta  
    = [(state5prime2,(updateQueue queue state5prime2),time-delta)]
--
-- 6' -- non-local transition, there is sft, but we cannot complete the transition, we must update queue
--
  | sft > 0 && sft < delta
    = [(state6prime,(updateQueue queue state6prime),time-delta)]
  where agent1       = transAgent tr
        n1           = agentNumber agent1
        sft          = agentSFT agent1 state
        trTime       = transDuration tr
        state5prime  = updateAllTimers delta (addContextInfo n2 (CLock agent1) (addContextInfo n1 (CSFT (trTime-delta)) state))
        state5prime2 = updateAllTimers delta (addContextInfo n1 (CSFT (trTime-delta)) state)
        state6prime  = updateAllTimers delta (addContextInfo n1 (CSFT (sft-delta)) (delContextInfo n1 (CSFT sft) state))
        agent2       = sndAgent tr
        n2           = agentNumber agent2
   
---------------------------------------------------------------------

-- | Function generates the LTS graph. It starts with initial
--   state as the only node of the LTS graph and the only node to proceed
--   and then builds it recursively.

ltsgen :: [Node]          -- ^ List of unprocessed nodes. We focus on the first node in the list.
       -> [TTransition]   -- ^ List of transitions enable in the first state of the unprocessed nodes.
       -> [Node]          -- ^ Auxiliary list of already generated nodes.
       -> [Node]          -- ^ Final LTS graph.

ltsgen unprocNodes@((number,state,queue,time,arcs):nodes) (tr:trs) auxList = ltsgen newUnprocNodes trs newAuxList
  where shift = timeShift state tr time
        results = timeFire shift tr state queue time
        (newUnprocNodes,newAuxList) = update (unprocNodes,auxList) tr shift results

-- Omit the first node if all transitions have been considered.
ltsgen (node1 : node2@(number,state,queue,time,arcs) : nodes) [] auxList
  = node1 : ltsgen (node2:nodes) (alpha1EnableTransitions state queue time) auxList

ltsgen [node] [] _ = [node]   -- We finish if there is no enable transitions. 


-- | LTS graph.
lts = ltsgen [(0,s0,q0,t0,[])] (alpha1EnableTransitions s0 q0 t0) [(0,s0,q0,t0,[])]
--lts = ltsgen [(0,s0,q0,0,[])] (alpha1EnableTransitions s0 q0 0) [(0,s0,q0,0,[])]

----------------------------------------------------------------------

{-|
    Udpates the list of unprocessed nodes and the list of auxiliary
    nodes. For each state from the list of new states, the function
    checks whether the auxiliary list (list of all already generated
    nodes) already contains a node with such a state. If so, only new
    arc is added to the first node of unprocessed nodes. Otherwise,
    a new node is added to both lists (and the corresponding arc to
    the first node of unprocessed nodes).
-}
update :: ([Node], [Node])              -- ^ list of nodes to process and auxiliary list of already generated nodes
       -> TTransition                   -- ^ fired transition (used as the label for arsc)
       -> Int                           -- ^ time shift 
       -> [(State, [(Int,Int)], Int)]   -- ^ list of states to be added to both lists.
       -> ([Node], [Node])              -- ^ updated lists of nodes

update (unprocNodes@((number,state,queue,time,arcs):nodes), auxList) tr delta (newState@(s,q,t):newStates)

  | index == -1 = update (((number,state,queue,time,arcs ++ [(tr,delta,lastNumber+1)]):nodes) ++ [(lastNumber+1,s,q,t,[])],
                          auxList ++ [(lastNumber+1,s,q,t,[])]) tr delta newStates
  | otherwise = update ((number,state,queue,time,arcs ++ [(tr,delta,index)]):nodes,auxList) tr delta newStates
  where index = findState newState auxList
        (lastNumber,_,_,_,_) = last unprocNodes

update lists _ _ [] = lists   

----------------------------------------------------------------------

-- | Provides the number of the LTS node that contains given state.
findState :: (State, [(Int,Int)], Int)   -- ^ model state
          -> [Node]                      -- ^ list of nodes of the LTS graph
          -> Int                         -- ^ return value
findState _ [] = -1
findState (s,q,t) ((number,state,queue,time,arcs):nodes) | s == state && q == queue && t == time = number
                                                         | otherwise  = findState (s,q,t) nodes

----------------------------------------------------------------------

-- | Inserts list of context entries into the given set.
insertMany :: [ContextInfo] -> Set ContextInfo -> Set ContextInfo
insertMany [] set = set
insertMany entries set = union set (fromList entries)

----------------------------------------------------------------------

-- | Removes list of context entries from the given set.
deleteMany :: [ContextInfo] -> Set ContextInfo -> Set ContextInfo
deleteMany [] set = set
deleteMany entries set = difference set (fromList entries)

----------------------------------------------------------------------
