
module Main where

import Data.Set as Set

----------------------------------------------------------------------

-- | Takes successors numbers from te list of outgoing arcs for a time LTS graph.
succNumbers :: [([TTransition], Int, Int)] -> [Int]
succNumbers [] = []
succNumbers ((_,_,no):arcs) = no : (succNumbers arcs)

----------------------------------------------------------------------

-- | Generates String representation of parameters' values for all agents.
state2pv :: State -> [String]
state2pv state = dumpPVState 1 state

----------------------------------------------------------------------

-- | Generates String representation of parameters' values for all agents.
--   Auxiliary recursive function.
dumpPVState :: Int -> State -> [String]
dumpPVState number state 
  | number > modelSize = []
  | otherwise = (takepvString number state) : dumpPVState (number + 1) state

----------------------------------------------------------------------

-- | Generates list of successors numbers from the list of outgoing arcs for time LTS graph.
trans2numbers :: [([TTransition],Int,Int)] -> [Int]
trans2numbers [] = []
trans2numbers ((_,_,no):trans) = no : trans2numbers trans

----------------------------------------------------------------------

-- | Type that represents a node of the Data Graph.
-- (DG_node_number, parameters values, [numbers of successors],
-- [numbers of elements of equivalence class]), [numbers of predecessors],
-- number of successors, number of predecessors)

type DGNode = (Int, [String],[Int],[Int],[Int],Int,Int)

----------------------------------------------------------------------

-- | Generates a projection of list of Nodes into lst of DGNodes.
--   Information about parameters' values is stored in nodes only, 
--   and labes of arcs are omitted.
pvProjection :: [Node] -> [DGNode]
pvProjection [] = []
pvProjection ((no,state,trans):nodes)
  = (no,state2pv state,trans2numbers trans,[no],[],0,0) : pvProjection nodes

----------------------------------------------------------------------

-- | Returns the order number of the given node in the list of nodes.
--   The function takes the number (identifier) of a data graph node.
ordNumber :: Int -> [DGNode] -> Int
ordNumber n graph = ordNumberFromK n 0 graph

ordNumberFromK :: Int -> Int -> [DGNode] -> Int
ordNumberFromK n k ((no,_,_,_,_,_,_):xs)
  | n == no = k
  | otherwise = ordNumberFromK n (k+1) xs
ordNumberFromK _ _ [] = -1

----------------------------------------------------------------------

-- | Generates lists of predecessors for nodes of the given data graph.
addPredecessors :: [DGNode]   -- ^ auxiliary list of nodes for recursion
                -> [DGNode]   -- ^ current version of the generated list of nodes
                -> [DGNode]   -- ^ data graph with lists of predecessors
addPredecessors [] ys = ys
addPredecessors ((number,_,succs,_,_,_,_):xs) ys 
  = addPredecessors xs (addPredecessor number succs ys)  -- adds the "number" to list of predecessors
                                                         -- of any node from the succs list

----------------------------------------------------------------------

-- | Inserts p to the  list of predecessors of any node from the ns list.
addPredecessor :: Int         -- ^ number (identifier) of a node
               -> [Int]       -- ^ list of successors of the p node
               -> [DGNode]    -- ^ auxiliary list of nodes for recursion
               -> [DGNode]    -- ^ updated data graph
addPredecessor _ _ [] = []
addPredecessor p ns (node@(number,state,succs,classNodes,preds,succNo,predNo):nodes)
  | elem number ns = (number,state,succs,classNodes,p : preds,succNo,predNo) : addPredecessor p ns nodes
  | otherwise = node : addPredecessor p ns nodes

----------------------------------------------------------------------

-- | Inserts information about the numer of predecessors and successors to data graph nodes.
addSize :: [DGNode] -> [DGNode]
addSize [] = []
addSize ((number,state,succs,classNodes,preds,_,_):nodes)
  = (number,state,succs,classNodes,preds,length succs,length preds) : addSize nodes

----------------------------------------------------------------------

-- | Removes the n-th element from the list.
removeItem :: Int -> [a] -> [a]
removeItem n xs = (take n xs) ++ (Prelude.drop (n+1) xs)

----------------------------------------------------------------------

-- | Removes duplicates from the given list.
removeDuplicates :: (Eq a) => [a] -> [a]
removeDuplicates []     = []
removeDuplicates (x:xs) = x : removeDuplicates (removeElem x xs)

----------------------------------------------------------------------

-- | Replaces the "old" value with the "new" one.
updateList :: (Eq a) => a -> a -> [a] -> [a]
updateList _ _ [] = []
updateList new old (x:xs)
  | x == old  = new : xs
  | otherwise = x : updateList new old xs

----------------------------------------------------------------------

-- | Concatenets two lists. Removes duplicates.
listUnion :: (Eq a) => [a] -> [a] -> [a]
listUnion xs [] = xs
listUnion [] ys = ys
listUnion xs (y:ys)
  | elem y xs = listUnion xs ys
  | otherwise = listUnion (xs ++ [y]) ys

----------------------------------------------------------------------

-- | Updates edges of the data graph. Replaces the "old" node (identifier)
--   with the "new" one on lists of successors and predecessors.
updateEdges :: Int -> Int -> [DGNode] -> [DGNode]
updateEdges _ _ [] = []
updateEdges new old ((number,state,succs,classNodes,preds,succNo,predNo):nodes)
  = (number,state,updateList new old succs,classNodes,updateList new old preds,succNo,predNo) : updateEdges new old nodes

----------------------------------------------------------------------

-- | Removes duplicated arcs from the data graph.
removeMultipleArcs :: [DGNode] -> [DGNode]
removeMultipleArcs [] = []
removeMultipleArcs ((number,state,succs,classNodes,preds,succNo,predNo):nodes)
  = (number,state,removeDuplicates succs,classNodes,removeDuplicates preds,succNo,predNo) : removeMultipleArcs nodes

----------------------------------------------------------------------

-- | For a primary "x" node (n) returns the second node for the forward moving
--   reduction. If there is no such a node returns -1. 
nodeForForwardMoving :: Int -> [DGNode] -> Int
nodeForForwardMoving n graph       
  | succNo /= 1 = -1
  | succNo == 1 && predNo > 0 && nextNumber /= number && state == nextState && nextSuccNo > 0 = nextNumber
  | succNo == 1 && predNo == 0 && nextNumber /= number && state == nextState && nextPredNo == 1 = nextNumber
  | otherwise   = -1
  where (number,state,succs,classNodes,preds,succNo,predNo) = graph !! n
        nextNumber = head succs
        nextOrdNumber = ordNumber nextNumber graph
        (_,nextState,_,_,_,nextSuccNo,nextPredNo) = graph !! nextOrdNumber

----------------------------------------------------------------------

-- | For a primary "x" node (n) returns the second node for the backward moving
--   reduction. If there is no such a node returns -1.
nodeForBackwardMoving :: Int -> [DGNode] -> Int
nodeForBackwardMoving n graph
  | predNo /= 1 = -1
  | predNo == 1 && succNo > 0 && prevNumber /= number && state == prevState && prevPredNo > 0 = prevNumber
  | predNo == 1 && succNo == 0 && prevNumber /= number && state == prevState 
    && prevPredNo > 0 && prevSuccNo == 1 = prevNumber
  | otherwise   = -1
  where (number,state,succs,classNodes,preds,succNo,predNo) = graph !! n
        prevNumber = head preds
        prevOrdNumber = ordNumber prevNumber graph
        (_,prevState,_,_,_,prevSuccNo,prevPredNo) = graph !! prevOrdNumber

----------------------------------------------------------------------

-- | Implementation of the Forward Moving reduction.
forwardMoving :: Int         -- ^ first "x" node (primary node)
              -> Int         -- ^ second "x" node
              -> [DGNode]    -- ^ input data graph
              -> [DGNode]    -- ^ reduced data graph
forwardMoving n k graph
  = addSize ((take newNOrdNumber updatedGraph) 
    ++ [(nNumber,nState,newSuccs,nClassNodes ++ kClassNodes,newPreds,0,0)] 
    ++ (Prelude.drop (newNOrdNumber + 1) updatedGraph))
  where nOrdNumber = ordNumber n graph
        kOrdNumber = ordNumber k graph
        (nNumber,nState,_,nClassNodes,nPreds,_,_) = graph !! nOrdNumber
        (kNumber,_,kSuccs,kClassNodes,kPreds,kSuccNo,_) = graph !! kOrdNumber
        reducedGraph = removeItem kOrdNumber graph
        updatedGraph = removeMultipleArcs (updateEdges n k reducedGraph)
        newNOrdNumber = ordNumber n updatedGraph
        newPreds = updateList n k (listUnion nPreds (removeElem n kPreds))
        newSuccs = removeDuplicates (updateList n k kSuccs)

----------------------------------------------------------------------

-- | Implementation of the Backward Moving reduction.
backwardMoving :: Int         -- ^ first "x" node (primary node)
               -> Int         -- ^ second "x" node
               -> [DGNode]    -- ^ input data graph
               -> [DGNode]    -- ^ reduced data graph
backwardMoving n k graph
  = addSize ((take newNOrdNumber updatedGraph) 
    ++ [(nNumber,nState,newSuccs,nClassNodes ++ kClassNodes,newPreds,0,0)] 
    ++ (Prelude.drop (newNOrdNumber + 1) updatedGraph))
  where nOrdNumber = ordNumber n graph
        kOrdNumber = ordNumber k graph
        (nNumber,nState,nSuccs,nClassNodes,_,_,_) = graph !! nOrdNumber
        (kNumber,_,kSuccs,kClassNodes,kPreds,_,kPredNo) = graph !! kOrdNumber
        reducedGraph = removeItem kOrdNumber graph
        updatedGraph = removeMultipleArcs (updateEdges n k reducedGraph)
        newNOrdNumber = ordNumber n updatedGraph
        newSuccs = updateList n k (listUnion nSuccs (removeElem n kSuccs))
        newPreds = removeDuplicates (updateList n k kPreds)

----------------------------------------------------------------------

-- | Returns True if the nodes y1 and y2 (must be successors of the same node)
--   satisfy conditions of the Forward Merging reduction. Returns False otherwise.
readyForForwardMerging :: Int -> Int -> [DGNode] -> Bool
readyForForwardMerging y1 y2 graph
  | state1 == state2 && predNo1 == 1 && predNo2 == 1 && succNo1 > 0 && succNo2 > 0 = True
  | state1 == state2 && predNo1 == 1 && predNo2 == 1 && succNo1 == 0 && succNo2 == 0 = True
  | otherwise = False
  where (_,state1,_,_,_,succNo1,predNo1) = graph !! (ordNumber y1 graph)
        (_,state2,_,_,_,succNo2,predNo2) = graph !! (ordNumber y2 graph)

----------------------------------------------------------------------

-- | Returns True if the nodes y1 and y2 (must be successors of the same node)
--   satisfy conditions of the Backward Merging reduction. Returns False otherwise.
readyForBackwardMerging :: Int -> Int -> [DGNode] -> Bool
readyForBackwardMerging y1 y2 graph
  | state1 == state2 && succNo1 == 1 && succNo2 == 1 && predNo1 > 0 && predNo2 > 0 = True
  | otherwise = False
  where (_,state1,_,_,_,succNo1,predNo1) = graph !! (ordNumber y1 graph)
        (_,state2,_,_,_,succNo2,predNo2) = graph !! (ordNumber y2 graph)

----------------------------------------------------------------------

-- | For a primary "x" node (n) returns the pair of nodes for the forward merging
--   reduction. If there is no such a pair returns (-1,-1).
nodesForForwardMerging :: Int -> [DGNode] -> (Int,Int)
nodesForForwardMerging n graph
  | succNo < 2  = (-1,-1)
  | pairs /= [] = head pairs
  | otherwise   = (-1,-1)
  where (number,state,succs,classNodes,preds,succNo,predNo) = graph !! n
        pairs = [(y1,y2) | y1 <- succs, y2 <- succs, y1 /= y2, y1 /= number, y2 /= number, readyForForwardMerging y1 y2 graph]

----------------------------------------------------------------------

-- | For a primary "x" node (n) returns the pair of nodes for the backward merging
--   reduction. If there is no such a pair returns (-1,-1).
nodesForBackwardMerging :: Int -> [DGNode] -> (Int,Int)
nodesForBackwardMerging n graph
  | predNo < 2  = (-1,-1)
  | pairs /= [] = head pairs
  | otherwise   = (-1,-1)
  where (number,state,succs,classNodes,preds,succNo,predNo) = graph !! n
        pairs = [(y1,y2) | y1 <- preds, y2 <- preds, y1 /= y2, y1 /= number, y2 /= number, readyForBackwardMerging y1 y2 graph]

----------------------------------------------------------------------

-- | Implementation of the Forward Merging reduction.
forwardMerging :: Int         -- ^ first "x" node (primary node)
               -> Int         -- ^ first "y" node
               -> Int         -- ^ second "y" node
               -> [DGNode]    -- ^ input data graph
               -> [DGNode]    -- ^ reduced data graph
forwardMerging n k1 k2 graph 
  | newNOrdNumber < newK1OrdNumber
    = addSize ((take newNOrdNumber updatedGraph) 
    ++ [(nNumber,nState,removeElem k2 nSuccs,nClassNodes,newNPreds,0,0)] 
    ++ (take (newK1OrdNumber - newNOrdNumber - 1) (Prelude.drop (newNOrdNumber + 1) updatedGraph))
    ++ [(k1Number,k1State,newK1Succs,k1ClassNodes ++ k2ClassNodes,k1Preds,0,0)]
    ++ (Prelude.drop (newK1OrdNumber + 1) updatedGraph))
  | newK1OrdNumber < newNOrdNumber
    = addSize ((take newK1OrdNumber updatedGraph)
    ++ [(k1Number,k1State,newK1Succs,k1ClassNodes ++ k2ClassNodes,k1Preds,0,0)]
    ++ (take (newNOrdNumber - newK1OrdNumber - 1) (Prelude.drop (newK1OrdNumber + 1) updatedGraph))
    ++ [(nNumber,nState,removeElem k2 nSuccs,nClassNodes,newNPreds,0,0)] 
    ++ (Prelude.drop (newNOrdNumber + 1) updatedGraph))
  where nOrdNumber = ordNumber n graph
        k1OrdNumber = ordNumber k1 graph
        k2OrdNumber = ordNumber k2 graph
        (nNumber,nState,nSuccs,nClassNodes,nPreds,nSuccNo,nPredNo) = graph !! nOrdNumber
        (k1Number,k1State,k1Succs,k1ClassNodes,k1Preds,k1SuccNo,k1PredNo) = graph !! k1OrdNumber
        (_,_,k2Succs,k2ClassNodes,_,_,_) = graph !! k2OrdNumber
        reducedGraph = removeItem k2OrdNumber graph
        updatedGraph = removeMultipleArcs (updateEdges k1 k2 reducedGraph)
        newNOrdNumber = ordNumber n updatedGraph
        newK1OrdNumber = ordNumber k1 updatedGraph
        newK1Succs = listUnion k1Succs k2Succs
        newNPreds = removeDuplicates (updateList k1 k2 nPreds)

----------------------------------------------------------------------

-- | Implementation of the Forward Merging reduction.
backwardMerging :: Int         -- ^ first "x" node (primary node)
                -> Int         -- ^ first "y" node
                -> Int         -- ^ second "y" node
                -> [DGNode]    -- ^ input data graph
                -> [DGNode]    -- ^ reduced data graph
backwardMerging n k1 k2 graph
  | newNOrdNumber < newK1OrdNumber
    = addSize ((take newNOrdNumber updatedGraph) 
    ++ [(nNumber,nState,newNSuccs,nClassNodes,removeElem k2 nPreds,0,0)] 
    ++ (take (newK1OrdNumber - newNOrdNumber - 1) (Prelude.drop (newNOrdNumber + 1) updatedGraph))
    ++ [(k1Number,k1State,k1Succs,k1ClassNodes ++ k2ClassNodes,newK1Preds,0,0)]
    ++ (Prelude.drop (newK1OrdNumber + 1) updatedGraph))
  | newK1OrdNumber < newNOrdNumber
    = addSize ((take newK1OrdNumber updatedGraph)
    ++ [(k1Number,k1State,k1Succs,k1ClassNodes ++ k2ClassNodes,newK1Preds,0,0)]
    ++ (take (newNOrdNumber - newK1OrdNumber - 1) (Prelude.drop (newK1OrdNumber + 1) updatedGraph))
    ++ [(nNumber,nState,newNSuccs,nClassNodes,removeElem k2 nPreds,0,0)] 
    ++ (Prelude.drop (newNOrdNumber + 1) updatedGraph))
  where nOrdNumber = ordNumber n graph
        k1OrdNumber = ordNumber k1 graph
        k2OrdNumber = ordNumber k2 graph
        (nNumber,nState,nSuccs,nClassNodes,nPreds,nSuccNo,nPredNo) = graph !! nOrdNumber
        (k1Number,k1State,k1Succs,k1ClassNodes,k1Preds,k1SuccNo,k1PredNo) = graph !! k1OrdNumber
        (_,_,_,k2ClassNodes,k2Preds,_,_) = graph !! k2OrdNumber
        reducedGraph = removeItem k2OrdNumber graph
        updatedGraph = removeMultipleArcs (updateEdges k1 k2 reducedGraph)
        newNOrdNumber = ordNumber n updatedGraph
        newK1OrdNumber = ordNumber k1 updatedGraph
        newK1Preds = listUnion k1Preds k2Preds
        newNSuccs = removeDuplicates (updateList k1 k2 nSuccs)

----------------------------------------------------------------------

-- | Appplies reduction rules to the givem data graph. Reductions are applied
--   as long as at least one of them can be used. The first parameter is used
--   for recursion purposes.
graphReduction :: Int -> [DGNode] -> [DGNode]
graphReduction n graph 
  | n == length graph = graph
  | forwardNode /= -1 = graphReduction 0 (forwardMoving number forwardNode graph)
  | backwardNode /= -1 = graphReduction 0 (backwardMoving number backwardNode graph)
  | forwardPair /= (-1,-1) = graphReduction 0 (forwardMerging number (fst forwardPair) (snd forwardPair) graph)
  | backwardPair /= (-1,-1) = graphReduction 0 (backwardMerging number (fst backwardPair) (snd backwardPair) graph)
  | otherwise = graphReduction (n + 1) graph
  where (number,state,succs,_,preds,succNo,predNo) = graph !! n
        forwardNode = nodeForForwardMoving n graph
        backwardNode = nodeForBackwardMoving n graph
        forwardPair = nodesForForwardMerging n graph
        backwardPair = nodesForBackwardMerging n graph

----------------------------------------------------------------------

-- | Translates an LTS graph into Data Graph.
lts2dg :: [Node] -> [DGNode]
lts2dg graph = graphReduction 0 (addSize (addPredecessors initDGGraph initDGGraph))
  where initDGGraph = removeMultipleArcs (pvProjection graph)

----------------------------------------------------------------------

-- | Appplies moving reduction rules to the givem data graph. Reductions are applied
--   as long as at least one of them can be used. The first parameter is used
--   for recursion purposes.
movingGraphReduction :: Int -> [DGNode] -> [DGNode]
movingGraphReduction n graph 
  | n == length graph = graph
  | forwardNode /= -1 = graphReduction 0 (forwardMoving number forwardNode graph)
  | backwardNode /= -1 = graphReduction 0 (backwardMoving number backwardNode graph)
  | otherwise = graphReduction (n + 1) graph
  where (number,state,succs,_,preds,succNo,predNo) = graph !! n
        forwardNode = nodeForForwardMoving n graph
        backwardNode = nodeForBackwardMoving n graph

----------------------------------------------------------------------

-- | Prints the LTS graph using the DOT format.
printDOTDG :: [DGNode] -> String
printDOTDG dataGraph = "digraph G {\n\t rankdir=TB;\n" 
  ++ insertEscapeCharDG (dumpDOTNodesDG dataGraph) ++ dumpDOTTransDG dataGraph ++ "}\n"

----------------------------------------------------------------------

-- | Prints the data graph arcs using the DOT format.
dumpDOTTransDG :: [DGNode] -> String
dumpDOTTransDG [] = []
dumpDOTTransDG ((no,_,succs,_,_,_,_):nodes) = (dumpTransDG no succs) ++ dumpDOTTransDG nodes

----------------------------------------------------------------------

-- | Recursive function for printing a list of arcs for the given node.
dumpTransDG :: Int -> [Int] -> String
dumpTransDG _ [] = []
dumpTransDG source (no:nos) 
  = "\t" ++ (show source) ++ " -> " ++ (show no) ++ ";\n"
  ++ dumpTransDG source nos

----------------------------------------------------------------------

-- | Prints the data graph nodes using the DOT format.
dumpDOTNodesDG :: [DGNode] -> String
dumpDOTNodesDG [] = []
dumpDOTNodesDG ((number,state,_,classNodes,preds,noSucc,noPred):nodes)
  = "\t" ++ (show number)
  ++ " [label = \"" ++ "(" ++ (show number) ++ ")" ++ (dumpDOTStateDG 0 state)
  ++ "\\n" ++ (show  classNodes)
  ++ "\" shape=box];\n" ++ dumpDOTNodesDG nodes


-- dumpDOTNodesDG :: [DGNode] -> String
-- dumpDOTNodesDG [] = []
-- dumpDOTNodesDG ((number,state,_,classNodes,preds,noSucc,noPred):nodes)
--   = "\t" ++ (show number)
--   ++ " [label = \"" ++ "(" ++ (show number) ++ ")" ++ (dumpDOTStateDG 0 state)
--   ++ "\\n" ++ (show  classNodes)
--   ++ "\\npreds: " ++ (show  preds) ++ "\\n[" ++ (show  noSucc) ++ ", " ++ (show  noPred)
--   ++ "]\" shape=box];\n" ++ dumpDOTNodesDG nodes


----------------------------------------------------------------------

-- | Inserts escape characters to the string representaion of graph nodes (DOT format).
insertEscapeChar :: String -> String
insertEscapeChar (x:xs)
  | isPrefix "label = \"(" (x:xs) = "label = \"(" ++ (insertEscapeChar (Prelude.drop 10 (x:xs)))
  | isPrefix ")\" shape=box" (x:xs) = ")\" shape=box" ++ (insertEscapeChar (Prelude.drop 12 (x:xs)))
  | isPrefix "]\" shape=box" (x:xs) = "]\" shape=box" ++ (insertEscapeChar (Prelude.drop 12 (x:xs)))
  | x =='\"' = "\\\"" ++ (insertEscapeChar xs)
  | otherwise = x : (insertEscapeChar xs)
insertEscapeChar [] = []

----------------------------------------------------------------------

-- | Inserts escape characters to the string representaion of graph nodes (DOT format).
insertEscapeCharDG :: String -> String
insertEscapeCharDG (x:xs)
  | isPrefix "label = \"(" (x:xs) = "label = \"(" ++ (insertEscapeCharDG (Prelude.drop 10 (x:xs)))
  | isPrefix ")\" shape=box" (x:xs) = ")\" shape=box" ++ (insertEscapeCharDG (Prelude.drop 12 (x:xs)))
  | isPrefix "]\" shape=box" (x:xs) = "]\" shape=box" ++ (insertEscapeCharDG (Prelude.drop 12 (x:xs)))
  | x =='\"' = "\\\"" ++ (insertEscapeCharDG xs)
  | otherwise = x : (insertEscapeCharDG xs)
insertEscapeCharDG [] = []

----------------------------------------------------------------------

-- | Prints the state using the DOT format.
dumpDOTStateDG :: Int -> [String] -> String
dumpDOTStateDG _ [] = ""
dumpDOTStateDG number (x:xs) 
  | number >= modelSize = ""
  | otherwise = "\\n" ++ (show (agents !! number)) ++ ": " ++ x
  ++ dumpDOTStateDG (number + 1) xs

----------------------------------------------------------------------

main = do 
  putStrLn (printDOTDG (lts2dg lts))
--  putStrLn ((printDOT lts))

