{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE DeriveGeneric #-}
module Main where

import Data.Set as Set
import System.IO (hSetBuffering,stdout, BufferMode(..),hPutStrLn,stderr)
import System.Exit(exitSuccess)

import Network.Socket hiding (send, recv)
import Network.Socket.ByteString (send, recv)

import GHC.Generics

import Data.Aeson (ToJSON, FromJSON, eitherDecode', encode)
import Data.Binary.Get (runGetOrFail, getWord32be)
import Data.Text (Text, pack)
import Data.ByteString.Lazy (toStrict, fromStrict)
import qualified Data.ByteString as BS
import Data.ByteString.Builder (toLazyByteString, word32BE)
----------------------------------------------------------------------

data AgentState =
  AgentState {
    aName :: Text,
    aMode :: Text,
    programCounter :: Int,
    contextInfo :: Text,
    parameterValues :: Text
  } deriving (Show, Generic)

data Transition =
  Transition {
    ttype :: Text,
    ag :: Text,
    firstPort :: Text,
    secondPort :: Text,
    stmtNumber :: Int
  } deriving (Show, Generic)

data SimulationTrace =
  SimulationTrace {
    format :: Text,
    body :: Text
  } deriving (Show, Generic)

data Response =
  Response {
    actionType :: Text,
    ind :: Maybe Int
  } deriving (Show, Generic)

instance ToJSON AgentState
instance ToJSON Transition
instance ToJSON SimulationTrace

instance FromJSON Response

createAgentState' :: String -> String -> Int -> String -> String -> AgentState
createAgentState' name mode pc ci pv = AgentState {
  aName = pack name,
  aMode = pack mode,
  programCounter = pc,
  contextInfo = pack ci,
  parameterValues = pack pv
}

createTransition' :: String -> (String, String, String, Int) -> Transition
createTransition' t (ag, p1, p2, stmt) = Transition {
  ttype = pack t,
  ag = pack ag,
  firstPort = pack p1,
  secondPort = pack p2,
  stmtNumber = stmt
}

createSimulationTrace' :: String -> String -> SimulationTrace
createSimulationTrace' f bd = SimulationTrace {
  format = pack f,
  body = pack bd
}


transitionType :: TTransition -> String
transitionType TDelay{}     = "delay"
transitionType TExec{}      = "exec"
transitionType TExit{}      = "exit"
transitionType TIf{}        = "if"
transitionType TIn{}        = "in"
transitionType TInAP{}      = "in"
transitionType TInPP{}      = "in"
transitionType TInF{}       = "in"
transitionType TJump{}      = "jump"
transitionType TLoop{}      = "loop"
transitionType TLoopEvery{} = "loop"
transitionType TNull{}      = "null"
transitionType TOut{}       = "out"
transitionType TOutAP{}     = "out"
transitionType TOutPP{}     = "out"
transitionType TOutF{}      = "out"
transitionType TSelect{}    = "select"
transitionType TStart{}     = "start"
transitionType STInAP{}     = "wakeup"
transitionType STInPP{}     = "wakeup"
transitionType STOutAP{}    = "wakeup"
transitionType STOutPP{}    = "wakeup"
transitionType STDelayEnd{} = "timeout"
transitionType STLoopEnd{}  = "timeout"
transitionType STInEnd{}    = "timeout"
transitionType STOutEnd{}   = "timeout"
transitionType STTime{}     = "time"
transitionType TNone        = "none"


transitionParameters :: TTransition -> (String, String, String, Int)
transitionParameters (TDelay a n)        = (show a,"","",n)
transitionParameters (TExec a n)         = (show a,"","",n)
transitionParameters (TExit a n)         = (show a,"","",n)
transitionParameters (TIf a n)           = (show a,"","",n)
transitionParameters t@(TIn p n)         = (show $ transAgent t,show p,"",n)
transitionParameters t@(TInAP p1 p2 n)   = (show $ transAgent t,show p1,show p2,n)
transitionParameters t@(TInPP p1 p2 n)   = (show $ transAgent t,show p1,show p2,n)
transitionParameters t@(TInF p1 p2 n)    = (show $ transAgent t,show p1,show p2,n)
transitionParameters (TJump a n)         = (show a,"","",n)
transitionParameters (TLoop a n)         = (show a,"","",n)
transitionParameters (TLoopEvery a n)    = (show a,"","",n)
transitionParameters (TNull a n)         = (show a,"","",n)
transitionParameters t@(TOut p n)        = (show $ transAgent t,show p,"",n)
transitionParameters t@(TOutAP p1 p2 n)  = (show $ transAgent t,show p1,show p2,n)
transitionParameters t@(TOutPP p1 p2 n)  = (show $ transAgent t,show p1,show p2,n)
transitionParameters t@(TOutF p1 p2 n)   = (show $ transAgent t,show p1,show p2,n)
transitionParameters (TSelect a n)       = (show a,"","",n)
transitionParameters (TStart a n)        = (show a,"","",n)
transitionParameters t@(STInAP p1 p2 n)  = (show $ transAgent t,show p1,show p2,n)
transitionParameters t@(STInPP p1 p2 n)  = (show $ transAgent t,show p1,show p2,n)
transitionParameters t@(STOutAP p1 p2 n) = (show $ transAgent t,show p1,show p2,n)
transitionParameters t@(STOutPP p1 p2 n) = (show $ transAgent t,show p1,show p2,n)
transitionParameters (STDelayEnd a n)    = (show a,"","",n)
transitionParameters (STLoopEnd a n)     = (show a,"","",n)
transitionParameters t@(STInEnd p n)     = (show $ transAgent t,show p,"",n)
transitionParameters t@(STOutEnd p n)    = (show $ transAgent t,show p,"",n)
transitionParameters (STTime n)          = ("","","",n)
transitionParameters TNone               = ("","","",-1)


createAgents :: [Agent] -> BS.ByteString
createAgents = createByteString . Prelude.map show

createStates :: [State] -> BS.ByteString
createStates = createByteString . Prelude.map createProgramState

createProgramState :: State -> [AgentState]
createProgramState st = Prelude.map (\x -> createAgentState (takeam (x+1) st) (takepc (x+1) st) (takeci (x+1) st) (takepvString (x+1) st) (show (agents!!x))) [0..Prelude.length agents-1]

createAgentState :: Mode -> Int -> Set ContextInfo -> String -> String -> AgentState
createAgentState m pc ci pv name = createAgentState' name (show m) pc (show ci) pv

createTransitions :: [TTransition] -> BS.ByteString
createTransitions = createByteString . Prelude.map createTransition

createTransition :: TTransition -> Transition
createTransition t = createTransition' (transitionType t) (transitionParameters t)

createByteString :: (ToJSON a) => a -> BS.ByteString
createByteString = toStrict . encode
---------------------------------------------------------------------

sendMessage :: (Socket, SockAddr) -> BS.ByteString -> IO Int
sendMessage (sock, _) msg = send sock (BS.append (toStrict (toLazyByteString (word32BE (Prelude.fromIntegral (BS.length msg))))) msg)

getResponse :: (Socket, SockAddr) -> IO (Either String Response)
getResponse (sock, _) = do
  msgLengthString <- recv sock 4
  let maybeMsgLength = runGetOrFail getWord32be (fromStrict msgLengthString)
  case maybeMsgLength of
    Left (_,_,msg) -> return $ Left ("Parse message length error: " ++ msg)
    Right (_,_, msgLength) -> do
      msg <- recv sock (Prelude.fromIntegral msgLength)
      return $ eitherDecode' $ fromStrict msg

--TODO Use EitherT to get rid of nested case
getSelectedValue :: (Socket, SockAddr) -> ([State], [TTransition]) -> Int -> [a] -> IO (Either String a)
getSelectedValue conn history@(sHist, tHist) size lst = do
  resp <- getResponse conn
  case resp of
    Left msg -> return $ Left msg
    Right Response { actionType = action, ind = index } -> do
      case action of
        "trace" -> do
          _ <- sendMessage conn (generateTrace (reverse sHist, reverse tHist))
          getSelectedValue conn history size lst
        _ -> do
          case index of
            Nothing -> return $ Left "empty index"
            Just x -> do
              if x < 0 || x > size
                then return $ Left "Index out of bounds"
                else return $ Right (lst!!x)

generateTrace :: ([State], [TTransition]) -> BS.ByteString
generateTrace (states, transitions) = createByteString $ createSimulationTrace' "dot" (printDOT (generateLts states transitions))

applySelectedElement :: (c -> b) -> (c -> ([State], [TTransition])) -> Either String c -> IO (Either String (b,([State], [TTransition])))
applySelectedElement _ _ (Left msg) = return $ Left msg
applySelectedElement genAns modifyHistory (Right el) = return $ Right ((genAns el), (modifyHistory el))

sendOverSocket :: (Socket, SockAddr) -> (a -> [c]) -> ([c] -> BS.ByteString) -> (a -> c -> b) -> (([State], [TTransition]) -> c -> ([State], [TTransition])) -> Either String (a, ([State], [TTransition])) -> IO (Either String (b, ([State], [TTransition])))
sendOverSocket _ _ _ _ _ (Left msg) = return $ Left msg
sendOverSocket conn genLst conv genAns modifyHistory (Right (smth, progState)) = do
  let lst = genLst smth
  let maxLength = Prelude.length lst -1
  _ <- sendMessage conn (conv lst)
  getSelectedValue conn progState maxLength lst >>= (applySelectedElement (genAns smth) (modifyHistory progState))

simulationOverSocket :: (Socket, SockAddr) -> Either String ([State], ([State], [TTransition]))  -> IO ()
simulationOverSocket _ (Left msg) = putStrLn msg
simulationOverSocket conn progState  = do
  sendOverSocket conn (id) (createStates) (\_ x -> x) (\(sHist, tHist) s -> (s:sHist, tHist)) progState
    >>= sendOverSocket conn (enableTransitions) (createTransitions) (flip fire) (\(sHist, tHist) t -> (sHist, t:tHist))
      >>= simulationOverSocket conn

serve :: Socket -> IO ()
serve sock = do
  conn <- accept sock
  simulationOverSocket conn (Right ([s0], ([],[])))

main = do
  sock <- socket AF_INET Stream 0
  setSocketOption sock ReuseAddr 1
  bind sock (SockAddrInet 9090 iNADDR_ANY)
  listen sock 1
  serve sock
  --simulation [s0]

----------------------------------------------------------------------

-- | Prints the state using the TXT format.
state2txt :: Int -> State -> String
state2txt number state
  | number > modelSize = ""
  | otherwise = (show (agents !! (number - 1))) ++ ": ("
  ++ (show (takeam number state)) ++ ","
  ++ (show (takepc number state)) ++ ","
  ++ (show (toList (takeci number state))) ++ ","
  ++ (takepvString number state) ++ ")  "
  ++ (state2txt (number + 1) state)

----------------------------------------------------------------------
