-- Priorities: P1 --

----------------------------------------------------------------------

-- | For a given transition function returns priority of agent
--   that is waiting for the availability of the port.
transPriority :: TTransition -> Int
transPriority tr = agentPriority (agentName (fstPort tr))

----------------------------------------------------------------------

-- | Returns the highest priority of agents that are in conflict.
highestPriority :: Port            -- ^ port
                -> [TTransition]   -- ^ input list of transitions
                -> Int             -- ^ return value
highestPriority _ [] = minPriority
highestPriority port (tr:trs)
  | ((isInOutTransition tr) && (port == (sndPort tr)))
    = min (transPriority tr) (highestPriority port trs)
  | otherwise = highestPriority port trs

----------------------------------------------------------------------

-- | Returns the transitions that are in conflict with the specified
--   transition and in which the agent priority is greater than
--   or equal to the priority specified in the argument.
transWithHighestPriority :: Port            -- ^ port
                         -> Int             -- ^ priority
                         -> [TTransition]   -- ^ input list of transitions
                         -> [TTransition]   -- ^ output list of transitions
transWithHighestPriority _ _ [] = []
transWithHighestPriority port priority (tr:trs)
  | isInOutTransition tr && transPriority tr <= priority && port == sndPort tr
    = tr : transWithHighestPriority port priority trs
  | otherwise = transWithHighestPriority port priority trs

----------------------------------------------------------------------

{-|
    Returns the set of transition enable in the current state.
    In contrust to the /enable/ function the /priorityEnable/ function takes
    priorities under consideration. The algorithm is as follows:
    (1) Look at the list of transitions. If this is a transition type of In or Out:
    Get from the input list transitions with the highest priority which are
    in the conflict with the considered transition. Attach them to the result list.
    Call the priorityEnable function recursively with a list that has removed
    all of the transitions considered in the previous step.
    (2) Otherwise add transition to the result list.
-}
priorityEnable :: [TTransition] -- ^ input list of transitions
               -> [TTransition] -- ^ output list of transitions
priorityEnable [] = []
priorityEnable (tr:trs)
  | isInOutTransition tr
    = concat [(transWithHighestPriority (sndPort tr) (highestPriority (sndPort tr) (tr:trs)) (tr:trs)),
    (priorityEnable (removeTransitions2 (sndPort tr) trs))]
  | otherwise = tr : (priorityEnable trs)

----------------------------------------------------------------------

-- | Provides the list of all transition enable in the given state.
enableTransitions :: State -> [TTransition]
enableTransitions state = priorityEnable (enableInState state)

----------------------------------------------------------------------
